from pmoired import oimodels, dpfit
import numpy as np
import matplotlib.pyplot as plt

def varVsErr(y, e, x=None, n='auto', verbose=0, fig=None, normalised=False):
    """
    compare the variance in a data vector 'y' (as function of optional 'x') with error 'e':
    perform a polynomial fit of order 'n' (default 2) and compare the variance of the residuals
    to the errors and deduce the correlations

    fig="int" will show a plot

    returns {'rho':correlation, 'err':median error} -> can be used for building 'correlations'
    dictionnary for dpfit.leastsqFit

    """
    if x is None:
        x = np.linspace(-1, 1, len(y))
    ptpx = np.ptp(x)
    if ptpx==0:
        ptpx = 1
    if type(n)==int:
        p = {'A%d'%i:0.1*np.ptp(y)*np.random.rand()/ptpx**i for i in range(n+1)}
        p['A0'] = np.mean(y)
        fit = dpfit.leastsqFit(dpfit.polyN, x-np.mean(x), p, y, e, verbose=0)
    elif n=='auto':
        j=1
        test = True
        chi2 = 0
        while test:
            p = {'A%d'%i:0.1*np.ptp(y)*np.random.rand()/ptpx**i for i in range(j+1)}
            p['A0'] = np.mean(y)
            fit = dpfit.leastsqFit(dpfit.polyN, x-np.mean(x), p, y, e, verbose=0)
            if chi2==0:
                chi2 = fit['chi2']
                test = j<len(y)//2-1
            else:
                test = (chi2-fit['chi2'])/chi2 > 0.01 and j<(len(y)//2-1)
                chi2 = fit['chi2']
            j+=1
        n = j-1

    # -- how to justify this? was done numerically...
    rho = 1 - np.std(y-fit['model'])**2/np.median(e)**2
    rho = min(rho,1)

    if rho<=0: # -> means data variance is same as uncertainty
        if normalised:
            res = {'rho':0, 'err':np.std(y-fit['model'])/np.median(e),
                   'poly':fit['best'], 'x0':np.mean(x)}
        else:
            res = {'rho':0, 'err':np.std(y-fit['model']),
                   'poly':fit['best'], 'x0':np.mean(x)}
    else:
        if normalised:
            res = {'rho':rho, 'err':1,
                   'poly':fit['best'], 'x0':np.mean(x)}
        else:
            res = {'rho':rho, 'err': np.median(e),
                   'poly':fit['best'], 'x0':np.mean(x)}

    # -- check
    if not fig is None:
        verbose=True
    if verbose:
        corr = {'catg':np.ones(len(y)),
                'rho':{1:res['rho']},
                'err':{1:res['err']},
               }
        fitc = dpfit.leastsqFit(dpfit.polyN, x-np.mean(x), p, y, e, verbose=0,
                                        correlations=corr)
        if fig:
            fit = dpfit.randomParam(fit, N=100)
            fitc = dpfit.randomParam(fitc, N=100)
            s = r'$\chi^2$ %.3f -> %.3f (with $\rho$=%.3f and [e]=%.4f)'%(fit['chi2'], fitc['chi2'],
                                                                          res['rho'], res['err'])
            dx = 0.2*np.mean(np.gradient(x))
            plt.close(fig)
            plt.figure(fig)
            plt.suptitle(s)
            ax1 = plt.subplot(211)
            res['ax1'] = ax1

            plt.plot(x, y, '.k')
            plt.errorbar(x, y, yerr=e, linestyle='none', color='k', capsize=2)
            plt.errorbar(x+dx, y, yerr=res['err'], linestyle='none', color='b', capsize=2, alpha=0.2)

            plt.fill_between(x, fit['r_ym1s'], fit['r_yp1s'], color='r', alpha=0.2)
            plt.fill_between(x, fitc['r_ym1s'], fitc['r_yp1s'], color='c', alpha=0.2)

            ax2 = plt.subplot(212, sharex=ax1)
            res['ax2'] = ax2

            plt.plot(x, y-fitc['model'], '.k')
            plt.errorbar(x, y-fitc['model'], yerr=e, linestyle='none', color='k', capsize=2)
            plt.errorbar(x+dx, y-fitc['model'], yerr=res['err'], linestyle='none', color='b', capsize=2, alpha=0.2)

            plt.fill_between(x, fit['r_ym1s']-fitc['model'], fit['r_yp1s']-fitc['model'], color='r', alpha=0.2)
            plt.fill_between(x, fitc['r_ym1s']-fitc['model'], fitc['r_yp1s']-fitc['model'], color='c', alpha=0.2)
            plt.tight_layout()
        else:
            s = 'chi2 %.3f -> %.3f (with rho=%.3f and [e]=%.4f)'%(fit['chi2'], fitc['chi2'],
                                                                  res['rho'], res['err'])
            print(s)

    return res

def corrSpectra(res):
    """
    takes results from oimodels.residualsOI(..., fullOutput=True) and compute correlations per spectra.

    return a
    """
    resi, wh, wl, data, err, models, ins = res
    _wh = np.array(wh)
    corr = {'catg':_wh, 'rho':{}, 'err':{}, 'poly':{}, 'x0':{}}
    for T in ['V2', '|V|', 'T3PHI']:
        tags = set(filter(lambda x: x.startswith(T), wh))
        #print(T, tags)
        for t in tags:
            w = np.where(_wh==t)
            tmp = varVsErr(data[w], err[w], wl[w], normalised=False)
            if tmp['rho']>0:
                corr['rho'][t] = float(tmp['rho'])
                corr['err'][t] = float(tmp['err'])
                corr['poly'][t] = tmp['poly']
                corr['x0'][t] = tmp['x0']
    return corr



# check transfer function
import pmoired.oifits as oifits

from matplotlib import pyplot as plt
import numpy as np


def showTF(files, insname=None, mode=None, wlmin=None, wlmax=None):
    """
    mode as function of insname:

    GRAVITY_SC: ('ESO INS SPEC RES', 'ESO INS POLA MODE', 'ESO DET2 SEQ1 DIT')
        e.g. ('MEDIUM', 'COMBINED', 3.0)
    GRAVITY_FT: ('ESO FT POLA MODE')
        e.g. ('COMBINED')

    """
    print('loading...', end=' ')
    files = oifits._globlist(files)
    data = oifits.loadOI(files, insname=insname, withHeader=True, verbose=False)
    print('done')
    if insname is None:
        insname = set([d['insname'] for d in data])
        if len(insname)>1:
            print('more that on instrument!', insname)
            return
        else:
            insname = insname[0]

    if insname=='GRAVITY_SC':
        keys = {'target':'ESO OBS TARG NAME',
                'cal':'ESO PRO CATG', # contains "CAL" if a calibrator, otherwise not
                'mode':('ESO INS SPEC RES', 'ESO INS POLA MODE', 'ESO DET2 SEQ1 DIT'),
           }
        nT = 4
        obs = [('OI_VIS', '|V|'), ('OI_T3', 'T3PHI')]

    elif insname=='GRAVITY_FT':
        keys = {'target':'ESO OBS TARG NAME',
                'cal':'ESO PRO CATG', # contains "CAL" if a calibrator, otherwise not
                #'mode':('ESO FT POLA MODE', 'ESO FT MODE'),
                'mode':('ESO FT POLA MODE'),
                }
        obs = [('OI_VIS', '|V|'), ('OI_T3', 'T3PHI')]
        nT = 4
    else:
        print('I need rules for "insname"=', insname)

    def _mode(h):
        _m = []
        for m in keys['mode']:
            if m in h:
                tmp = h[m]
            else:
                tmp = 'no found'
            if type(tmp)==str:
                tmp = tmp.strip()
            _m.append(tmp)
        return tuple(_m)

    # == plots
    axV, pV = {}, 1
    axT, pT = {}, 3

    modes = [_mode(d['header']) for d in data]

    if len(set(modes))==1:
        dataM = data
    else:
        if mode is None:
            print('more than one mode!', set(modes))
            return
        # == select data according to mode ==
        dataM = [d for d in data if _mode(d['header'])==mode]

    MJD0 = int(min([min(d['MJD']) for d in dataM]))

    #plt.close(0)
    #figv = plt.figure(0, figsize=(10,6))
    #plt.close(1)
    #figt = plt.figure(1, figsize=(10,6))

    plt.close(0)
    plt.figure(0, figsize=(10,6))

    #wlmin, wlmax = 2.15, 2.2
    if wlmin is None:
        wlmin = max([min(d['WL']) for d in dataM])
    if wlmax is None:
        wlmax = min([max(d['WL']) for d in dataM])
    ic, dc = 0, {}
    color = ['r', 'g', 'b', 'orange', 'm', 'c', 'y']
    nc = 10 # number of colors to sample in the color map
    color = [plt.cm.tab10((i+0.5)/nc) for i in range(nc)]

    last_targ = ''
    for d in dataM:
        targ = d['header'][keys['target']]
        catg = d['header'][keys['cal']]
        if not targ in dc:
            dc[targ] = color[ic%len(color)]
            ic+=1
        # -- below specific to GRAVITY, needs to be generalised!
        # --  V --
        for k in d['OI_VIS'].keys():
            if not k in axV:
                if nT==4:
                    #axV[k] = figv.add_subplot(3,2,pV)
                    axV[k] = plt.subplot(3,4,pV)
                    if pV>=9:
                        axV[k].set_xlabel('MJD-%d'%MJD0)
                axV[k].set_title('|V| '+k, fontsize=8, x=0.2, y=0.8)
                pV+=1
                if nT==4 and pV in [3,7]:
                    pV+=2
            for i,V in enumerate(d['OI_VIS'][k]['|V|']): # for each baseline
                w = (d['WL']>=wlmin)*(d['WL']<=wlmax)
                if np.nanmean(V[w])==0:
                    m = 'x'
                    c = '0.5'
                else:
                    m = '^' if 'CAL' in catg else '*'
                    c = dc[targ]
                axV[k].plot(d['OI_VIS'][k]['MJD'][i]-MJD0, np.nanmean(V[w]), m, color=c)
                if targ!=last_targ:
                    axV[k].text(d['OI_VIS'][k]['MJD'][i]-MJD0, np.nanmean(V[w]),
                                targ, rotation=90, color=dc[targ], fontsize=6)
        # -- T3 --
        for k in d['OI_T3'].keys():
            if not k in axT:
                if nT==4:
                    #axT[k] = figt.add_subplot(2,2,pT)
                    axT[k] = plt.subplot(2,4,pT)
                    if pT>=7:
                        axT[k].set_xlabel('MJD-%d'%MJD0)

                axT[k].set_title('T3PHI '+k, fontsize=8, x=0.3, y=0.9)
                pT+=1
                if nT==4 and pT in [5]:
                    pT+=2

            for i,T3 in enumerate(d['OI_T3'][k]['T3PHI']): # for each triangle
                w = (d['WL']>=wlmin)*(d['WL']<=wlmax)
                if np.nanmean(T3[w])==0:
                    m = 'x'
                    c='0.5'
                else:
                    m = '^' if 'CAL' in catg else '*'
                    c = dc[targ]
                axT[k].plot(d['OI_T3'][k]['MJD'][i]-MJD0, np.nanmean(T3[w]), m, color=c)
                if targ!=last_targ:
                    axT[k].text(d['OI_T3'][k]['MJD'][i]-MJD0, np.nanmean(T3[w]),
                                targ, rotation=90, color=dc[targ], fontsize=6)
        if targ!=last_targ:
            last_targ = targ

    #plt.tight_layout()
    plt.subplots_adjust(wspace=0.4, hspace=0.3, top=.99)
    return



import numpy as np
from astroquery.simbad import Simbad
from matplotlib import pyplot as plt
import itertools

try:
    # Necessary while Python versions below 3.9 are supported.
    import importlib_resources as resources
except ImportError:
    from importlib import resources

import scipy.special

from pmoired import oimodels

# -- generate fake VLTI data

# -- some problems found by G. Duvert in the 'A0' terms (last col):
# @wvgvlti:Appl_data:VLTI:vltiParameter:telStations.nomCentre
# -- altitude: W=4.5397 fot ATs, W=13.044 for the UTs
vlti_latitude = -24.62743941
vlti_longitude = -70.40498688
layout_orientation = -18.984  # degrees
layout = {'A0': (-32.0010, -48.0130, -14.6416, -55.8116, 129.8495),
          'A1': (-32.0010, -64.0210, -9.4342, -70.9489, 150.8475),
          'B0': (-23.9910, -48.0190, -7.0653, -53.2116, 126.8355),
          'B1': (-23.9910, -64.0110, -1.8631, -68.3338, 142.8275),
          'B2': (-23.9910, -72.0110, 0.7394, -75.8987, 150.8275, ),
          'B3': (-23.9910, -80.0290, 3.3476, -83.4805, 158.8455),
          'B4': (-23.9910, -88.0130, 5.9449, -91.0303, 166.8295),
          'B5': (-23.9910, -96.0120, 8.5470, -98.5942, 174.8285),
          'C0': (-16.0020, -48.0130, 0.4872, -50.6071, 118.8405),
          'C1': (-16.0020, -64.0110, 5.6914, -65.7349, 134.8385),
          'C2': (-16.0020, -72.0190, 8.2964, -73.3074, 142.8465),
          'C3': (-16.0020, -80.0100, 10.8959, -80.8637, 150.8375),
          'D0': (0.0100, -48.0120, 15.6280, -45.3973, 97.8375),
          'D1': (0.0100, -80.0150, 26.0387, -75.6597, 134.8305),
          'D2': (0.0100, -96.0120, 31.2426, -90.7866, 150.8275),
          'E0': (16.0110, -48.0160, 30.7600, -40.1959, 81.8405),
          'G0': (32.0170, -48.0172, 45.8958, -34.9903, 65.8357),
          'G1': (32.0200, -112.0100, 66.7157, -95.5015, 129.8255),
          'G2': (31.9950, -24.0030, 38.0630, -12.2894, 73.0153),
          'H0': (64.0150, -48.0070, 76.1501, -24.5715, 58.1953),
          'I1': (72.0010, -87.9970, 96.7106, -59.7886, 111.1613),
          # -- XY are correct, A0 is guessed!
          'I2': (80, -24, 83.456, 3.330, 90),
          'J1': (88.0160, -71.9920, 106.6481, -39.4443, 111.1713),
          'J2': (88.0160, -96.0050, 114.4596, -62.1513, 135.1843),
          'J3': (88.0160, 7.9960, 80.6276, 36.1931, 124.4875),
          'J4': (88.0160, 23.9930, 75.4237, 51.3200, 140.4845),
          'J5': (88.0160, 47.9870, 67.6184, 74.0089, 164.4785),
          'J6': (88.0160, 71.9900, 59.8101, 96.7064, 188.4815),
          'K0': (96.0020, -48.0060, 106.3969, -14.1651, 90.1813),
          'L0': (104.0210, -47.9980, 113.9772, -11.5489, 103.1823),
          'M0': (112.0130, -48.0000, 121.5351, -8.9510, 111.1763),
          'U1': (-16.0000, -16.0000, -9.9249, -20.3346, 189.0572),
          'U2': (24.0000, 24.0000, 14.8873, 30.5019, 190.5572),
          'U3': (64.0000, 48.0000, 44.9044, 66.2087, 199.7447),
          'U4': (112.0000, 8.0000, 103.3058, 43.9989, 209.2302),
          # 'N0':[188,     -48,     0,        0, 188],
          # 'J7':[ 88,    -167,     0,        0, 200]
          }
# @wvgvlti:Appl_data:VLTI:vltiParameter:optics.delayLineParams
# --            U      Vin      Vout
DL_UV = {'DL1': (58.92, -37.245, -37.005),
         'DL2': (58.92, -37.995, -37.755),
         'DL3': (45.08, -38.505, -38.745),
         'DL4': (45.08, -39.255, -39.495),
         'DL5': (58.92, -40.745, -40.505),
         'DL6': (58.92, -41.495, -41.255)}

# @wvgvlti:Appl_data:VLTI:vltiParameter:optics.LabInputCoord
# -- U coordinates of IP 1 through 8:
IP_U = {'IP1': 52.32, 'IP2': 52.56, 'IP3': 52.80, 'IP4': 53.04,
        'IP5': 53.28, 'IP6': 53.52, 'IP7': 53.76, 'IP8': 54.00}

# @wvgvlti:Appl_data:VLTI:vltiParameter:optics.SwitchYardParams
# -- OPL per input: direct, BC, DDL, BC+DDL
OPL = {'IP1': (2.105, 11.5625, 11.5298, 20.9873),
       'IP2': (1.385, 12.7625,  9.6098, 20.9873),
       'IP3': (2.825, 11.8025, 11.2898, 20.2673),
       'IP4': (2.105, 13.0025,  9.3698, 20.2673),
       'IP5': (3.545, 12.0425, 11.0498, 19.5473),
       'IP6': (2.825, 13.2425,  9.1298, 19.5473),
       'IP7': (4.265, 12.2825, 10.8098, 18.8273),
       'IP8': (3.545, 13.4825,  8.8898, 18.8273), }

# ---------- horizons ---------------
# -- seems to come from VLT-TRE-ESO-15000-2551
# -- https://pdm.eso.org/kronodoc/1100/Get/265745/VLT-TRE-ESO-15000-2551_1.PDF


def loadHorizon(filename=None):
    # -- use custom computation hzlib.tabulateAll()
    with resources.as_file(resources.files('pmoired').joinpath('newVltiHzn_obsDoors.txt')) as p:
        if filename is None:
            filename = p
        with open(filename, 'r') as f:
            # print('- reading VLTI horizons from '+filename)
            for l in f.readlines():
                if not l.startswith('#'):
                    az.append(float(l.split()[0]))
                    for i, k in enumerate(keys):
                        horizon[k].append(float(l.split()[i+1]))
                elif l.startswith('# az'):
                    keys = l.split('# az')[1].split()
                    az = []
                    horizon = {k: [] for k in keys}
                else:
                    pass
    for k in keys:
        horizon[k] = (az, horizon[k])
    return horizon


horizon = loadHorizon()


def nTelescopes(T, target, lst, ip=None, DLconstraints=None,
                min_altitude=30, max_OPD=100, max_vcmPressure=None,
                max_airmass=None, flexible=True, plot=False, STS=True):
    """
    Compute u, v, alt, az, airmass, tracks for a nT configurartion

    - T: list of telescopes stations OR stationDL. example: ['A0','K1']
      or ['A0DL1','K0DL4']. If delaylines are given, it will compute the optimum
      position for the dlines and VCM pressure.

    - target: Can be a string with the name of the target resolvable by simbad.
      Can also be a tuple of 2 decimal numbers for the coordinates (ra, dec), in
      decimal hours and decimal degrees

    - lst: a list (or ndarray) of local sidearal times.

    - [ip]: list of input channels. The order should be the same as telescopes.

    - min_altitude: for observability computation, in degrees (default=20deg)

    - max_OPD: for observability, in m (default=100m)

    - max_vcmPressure: enforce the VCM limitation (SLOW!!!)

    - flexible: if False, qperform some checks on the Array Configuration: no
      AT/UT; no more than 4 ATs west of the I track (DL1,2,4,5) and no more
      than 2 ATs east ofq the G track (DL3,4); no more than one AT per track
      (e.g. A0-A1 is not possible). *flexible=False will allow currently
      impossible configurations*

    - plot: True/False

    - DLconstraints sets min and max. For example: {'min1':35, 'max2':70}

    Returns: return a self explanatory dictionnary. units are m and degrees

    the key 'observable' is a table of booleans (same length as 'lst')
    indicating wether or not the object is observable.

    """
    tmp = []
    if ip is None:
        # -- assumes 1,3,5,7 for 4T
        ip = range(2*len(T))[1::2]

    if all(['DL' in x for x in T]):
        dl = [int(t.split('DL')[1]) for t in T]
        T = [t.split('DL')[0] for t in T]
        if len(set(dl)) < len(dl):
            print('ERROR: some DL used more than once')
            return None
        if len(set(T)) < len(T):
            print('ERROR: some STATIONS used more than once')
            return None
    else:
        dl = None

    if not isinstance(DLconstraints, dict):
        DLconstraints = {}

    if not flexible:
        # -- check is we have at most 4 Tel p<30m and at most 2T p>30m
        W = list(filter(lambda t: layout[t][0] <= 35, T))
        E = list(filter(lambda t: layout[t][0] >= 35, T))
        assert len(W) <= 4, 'Too many stations West of the laboratory'
        assert len(E) <= 2, 'Too many stations East of the laboratory'

        # -- check the delay lines association
        if not dl is None:
            for k in range(len(T)):
                assert (layout[T[k]][0] <= 35 and dl[k] in [1, 2, 5, 6]) or \
                       (layout[T[k]][0] >= 35 and dl[k] in [3, 4]), \
                    'wrong delay line / station association'

        # -- check that all telescopes are UTs if any UTs
        tracks = [t[0] for t in T]
        if 'U' in tracks:
            assert len(
                set(tracks)) == 1, 'UTs and ATs cannot be mixed, use flexible=False'
        else:
            if not any(['J' in t for t in T]):
                assert len(set(tracks)) == len(T), \
                    'only one AT per track is allowed, use flexible=False'
            else:  # special case for J stations
                _T = list(filter(lambda t: not t.startswith('J'), T))
                _tracks = [t[0] for t in _T]
                assert len(set(_tracks)) == len(_T), \
                    'only one AT per track is allowed, use flexible=False'
                _T = list(filter(lambda t: t.startswith('J'), T))
                _s = list(filter(lambda t: t in ['J1', 'J2'], T))
                _n = list(filter(lambda t: t in ['J3', 'J4', 'J5', 'J6'], T))
                assert len(
                    _s) <= 1, 'only AT on J south track, use flexible=False'
                assert len(
                    _n) <= 1, 'only AT on J north track, use flexible=False'

    if isinstance(target, str):
        # s = simbad.query(target)
        # target_name=target
        # target = [s[0]['RA.h'], s[0]['DEC.d']]
        s = Simbad.query_object(target)
        try:
            ra = np.sum(np.float64(s['RA'][0].split())
                        * np.array([1, 1/60., 1/3600.]))
            dec = np.float64(s['DEC'][0].split())*np.array([1, 1/60., 1/3600.])
            dec = np.sum(np.abs(dec))*np.sign(dec[0])
        except:
            ra, dec = 0, 0
            print('WARNING: simbad query failed for', target)
            print(s)
        target_name = target
        target = ra, dec
    else:
        # assume [ra dec]
        if int(60*(target[0]-int(target[0]))) < 10:
            target_name = 'RA=%2d:0%1d' % (int(target[0]),
                                           int(60*(target[0]-int(target[0]))),)
        else:
            target_name = 'RA=%2d:%2d' % (int(target[0]),
                                          int(60*(target[0]-int(target[0]))),)
        if int(60*(abs(target[1])-int(abs(target[1])))) < 10:
            target_name += ' DEC=%3d:0%1d' % (int(target[1]),
                                              int(60*(abs(target[1]) -
                                                      int(abs(target[1])))),)
        else:
            target_name += ' DEC=%3d:%2d' % (int(target[1]),
                                             int(60*(abs(target[1]) -
                                                     int(abs(target[1])))),)
    res = {}
    if not dl is None:
        res['config'] = [T[k]+'DL'+str(dl[k])+'IP'+str(ip[k])
                         for k in range(len(ip))]
    else:
        res['config'] = [T[k]+'IP'+str(ip[k]) for k in range(len(ip))]

    for i in range(len(T)+1):
        for j in range(len(T))[i+1:]:
            tmp = projBaseline(T[i], T[j],
                               target, lst, ip1=ip[i], ip2=ip[j],
                               DL1=None if dl is None else dl[i],
                               DL2=None if dl is None else dl[j],
                               min_altitude=min_altitude,
                               max_airmass=max_airmass,
                               max_OPD=max_OPD)
            if not 'lst' in res.keys():
                # init
                for k in ['lst', 'airmass', 'observable',
                          'ra', 'dec', 'alt', 'az', 'parang',
                          'horizon']:
                    res[k] = tmp[k]
                res['baselines'] = [tmp['baselines']]
                for k in ['B', 'PA', 'u', 'v', 'opd', 'ground']:
                    res[k] = {}
            else:
                # update
                res['observable'] = res['observable']*tmp['observable']
                res['baselines'].append(tmp['baselines'])
                res['horizon'] = np.maximum(res['horizon'],
                                            tmp['horizon'])
            for k in ['B', 'PA', 'u', 'v', 'opd', 'ground']:
                res[k][tmp['baselines']] = tmp[k]

    # -- convert OPL into DL position in the lab:
    # -- @wvgvlti:Appl_data:VLTI:vltiParameter:optics:LabInputCoord
    x_M16 = 52.  # where the center of the DL tunnel
    dx_dl = 6.92  # DL start to M16 -> real number!
    y_lab = -28
    # -- coordinate of 0-OPD in the lab
    y_ip = {'IP1': y_lab+5*0.24, 'IP2': y_lab+1*0.24,
            'IP3': y_lab+6*0.24, 'IP4': y_lab+2*0.24,
            'IP5': y_lab+7*0.24, 'IP6': y_lab+3*0.24,
            'IP7': y_lab+8*0.24, 'IP8': y_lab+4*0.24, }
    x_ip = x_M16 - 5

    wdl = .3  # width of the DL (drawing only)
    def fy_M12(x): return y_lab - 1.8*x  # M12 y as fct of DL index
    def fx_IP(x): return x_M16 + 1*x - 4  # IP x as fct of IP index

    minDL_opl = 11.  # smallest OPL for each cart

    # -- compute DL position and VCM position
    if not dl is None:  # Delay lines are given, so compute VCM limitation
        res['vcm'] = {}
        res['dl'] = {}
        for d in dl:
            res['vcm'][d] = []
            res['dl'][d] = []
        # -- compute OPD dictionnary
        for k in range(len(res['lst'])):
            opdD = {}
            for i1, t1 in enumerate(T):
                for i2, t2 in enumerate(T):
                    maxO = max_OPD
                    if t1+t2 in res['opd'].keys():
                        opdD[(dl[i1], dl[i2])] = res['opd'][t1+t2][k]
            s = solveDLposition(dl, opdD, stations=T, STS=STS,
                                dlPosMin=minDL_opl/2.,
                                dlPosMax=(max_OPD+minDL_opl)/2.,
                                constraints=DLconstraints)
            for d in dl:
                if np.isnan(s[d]):
                    res['observable'][k] = False
                res['dl'][d].append(s[d])
                res['vcm'][d].append(s['vcm'+str(d)])

        for d in dl:
            res['dl'][d] = np.array(res['dl'][d])
            res['vcm'][d] = np.array(res['vcm'][d])

        if not max_vcmPressure is None:
            maxP = 0
            for d in dl:
                maxP = np.maximum(maxP, res['vcm'][d])
            res['observable'] = res['observable']*(maxP < max_vcmPressure)

        res['x_dl'] = {}
        res['y_dl'] = {}

        for i, d in enumerate(dl):
            if d == 3 or d == 4:
                res['x_dl'][d] = x_M16 - dx_dl - res['dl'][d]
            else:
                res['x_dl'][d] = x_M16 + dx_dl + res['dl'][d]
            res['y_dl'][d] = fy_M12(d)

    # -- plot observability
    if plot:
        colors = ['r', 'g', 'b', 'y', 'c', 'm']
        colors = [(1, 0, 0), (0, 1, 0), (0, 0, 1),
                  (1, 0.5, 0), (0, 1, 0.5), (0.5, 0, 1)]
        colors = [(0.3+c[0]*0.5,
                   0.3+c[1]*0.5,
                   0.3+c[2]*0.5) for c in colors]

        def where(key): return [res[key][k] for k in
                                range(len(res[key]))
                                if res['observable'][k]]

        def where2(key1, key2): return [res[key1][key2][k] for k in
                                        range(len(res['lst']))
                                        if res['observable'][k]]
        if isinstance(plot, int):
            fig = plot
        else:
            fig = 0
        if not dl is None:
            plt.figure(fig, figsize=(12, 7))
            plt.clf()
            plt.subplots_adjust(hspace=0.16, top=0.94, left=0.06,
                                wspace=0.15, right=0.98, bottom=0.01)
            ax1 = plt.subplot(221)
        else:
            plt.figure(fig, figsize=(7, 5))
            plt.clf()
            plt.subplots_adjust(wspace=0.07, top=0.90, left=0.08,
                                right=0.98, bottom=0.12)
            ax1 = plt.subplot(111)

        plt.plot(res['lst'], res['lst']*0 + min_altitude, '-',
                 linestyle='dashed', color='k', linewidth=3, alpha=0.4)

        plt.plot(res['lst'], res['horizon'],
                 color=(0.1, 0.2, 0.4), alpha=0.3,
                 label='UT shadow', linewidth=2)
        plt.fill_between(res['lst'], res['horizon']*0.0,
                         res['horizon'], color=(0.1, 0.2, 0.4),
                         alpha=0.5)
        plt.plot(res['lst'], res['alt'], '+', alpha=0.9,
                 color=(0.5, 0.1, 0.3), label='not obs.',
                 markersize=8)
        plt.plot(where('lst'), where('alt'), 'o', alpha=0.5,
                 color=(0.1, 0.5, 0.3), label='observable',
                 markersize=8)
        plt.legend(prop={'size': 9}, numpoints=1)
        plt.ylim(0, 92)
        plt.xlabel('lst (h)')
        plt.ylabel('altitude (deg)')
        plt.suptitle(target_name+' '+'-'.join(res['config']) +
                     (' [STS]' if STS else ''))
        # plt.hlines(min_altitude, plt.xlim()[0], plt.xlim()[1],
        #              color='r', linewidth=1, linestyle='dotted')
        # -- make LST labels to be in 0..24
        # xl = plt.xticks()
        # xl = [str(int(x)%24) for x in xl[0]]
        # ax1.set_xticklabels(xl)
        plt.grid()
        # -- DL cart position
        if dl is None:
            return

        # -- OPL (2xDL position) for each DL:
        ax3 = plt.subplot(222, sharex=ax1)
        for i, d in enumerate(dl):
            plt.plot(where('lst'), 2*np.array(where2('dl', d)),
                     '-', label='DL'+str(d),
                     alpha=0.7, color=colors[i], linewidth=3)
            # -- plt ref DL
            wr = np.where(np.diff(where2('dl', d)) == 0)
            plt.plot(np.array(where('lst'))[wr], 2*np.array(where2('dl', d))[wr],
                     alpha=1, color=colors[i], linewidth=5,
                     linestyle='dashed')

        plt.legend(loc='upper center', prop={'size': 10}, numpoints=1,
                   ncol=6)
        plt.xlabel('lst (h)')
        plt.ylabel('OPL (m)')
        plt.hlines([minDL_opl, max_OPD+minDL_opl], res['lst'].min(), res['lst'].max(),
                   linestyle='dashed', color='k', linewidth=3, alpha=0.4)
        plt.grid()
        plt.ylim(0, 130)

        # ========================
        # == TUNNEL ==============
        # ========================

        # markers:
        DLm_W = [(-0.03, 0.04), (-0.1, 0.04), (-0.12, 0.01), (-0.12, -0.01),
                 (-0.1, -0.04), (-0.1, -0.06), (0.1, -0.06), (0.1, -0.04),
                 (-0.03, -0.04), (-0.03, 0.04)]

        DLm_E = [(-d[0], d[1]) for d in DLm_W]

        ax2 = plt.subplot(212)
        plt.text(x_M16, y_lab+0.3, 'LAB', ha='center', va='bottom',
                 color='0.5')
        # -- delay lines rails and ticks:
        for d in [1, 2, 3, 4, 5, 6]:
            s = (-1)**(d == 3 or d == 4)
            # -- range
            plt.plot([x_M16 + s*dx_dl, x_M16 + s*dx_dl + s*60],
                     [fy_M12(d)-wdl/2, fy_M12(d)-wdl/2], '-k',
                     alpha=0.5)
            plt.text(x_M16 + s*dx_dl + s*62, fy_M12(d)-wdl/2,
                     'DL'+str(d), va='center', color='0.5', size=9,
                     ha='right' if s == -1 else 'left')
            # -- 10m ticks
            plt.plot(x_M16 + s*dx_dl + s*np.linspace(0, 60, 7),
                     fy_M12(d)-wdl/2 + np.zeros(7), '|k',
                     markersize=9, color='0.6')
            # -- 5m ticks
            plt.plot(x_M16 + s*dx_dl + s*np.linspace(0, 50, 6)+s*5,
                     fy_M12(d)-wdl/2 + np.zeros(6), '|k',
                     markersize=7, color='0.4')
            for x in np.linspace(0, 60, 7):
                plt.text(x_M16 + s*dx_dl + s*x, fy_M12(d)+0.6*wdl,
                         str(int(2*x)), size=6, va='bottom',
                         ha='center', color='k')
            # -- DL restrictions
            for k in DLconstraints.keys():
                if str(d) in k:
                    # print(d, k, DLconstraints[k])
                    if 'min' in k:
                        plt.fill_between(x_M16 + s*dx_dl +
                                         s*np.array([0, DLconstraints[k]/2]),
                                         fy_M12(d)-np.array([0.9, 0.9])*wdl,
                                         fy_M12(d)-np.array([0.1, 0.1])*wdl,
                                         color='k', hatch='////', alpha=0.3)
                        plt.text(x_M16 + s*dx_dl + s*DLconstraints[k]/2,
                                 fy_M12(d)-2.5*wdl, str(DLconstraints[k]),
                                 color='r', size=8, va='bottom', ha='center')
                    elif 'max' in k:
                        plt.fill_between(x_M16 + s*dx_dl +
                                         s*np.array([60, DLconstraints[k]/2]),
                                         fy_M12(d)-np.array([0.9, 0.9])*wdl,
                                         fy_M12(d)-np.array([0.1, 0.1])*wdl,
                                         color='k', hatch='////', alpha=0.3)
                        plt.text(x_M16 + s*dx_dl + s*DLconstraints[k]/2,
                                 fy_M12(d)-2.5*wdl, str(DLconstraints[k]),
                                 color='r', size=8, va='bottom', ha='center')

        # -- used range:
        for i, d in enumerate(dl):
            # print(d, np.min(where2('dl', d)), np.max(where2('dl', d)))
            s = (-1)**(d == 3 or d == 4)
            # -- DL range
            plt.plot([x_M16 + s*dx_dl + s*np.min(where2('dl', d)),
                     x_M16 + s*dx_dl + s*np.max(where2('dl', d))],
                     [fy_M12(d)-wdl/2, fy_M12(d)-wdl/2], '-',
                     color=colors[i], alpha=0.5, linewidth=8,
                     label='range %d' % (d))

        # -- light path: T -> M12 -> cart -> M16
        for i, t in enumerate(T):
            y_M12 = fy_M12(dl[i])
            x_IP = fx_IP(ip[i])
            s = (-1)**(dl[i] == 3 or dl[i] == 4)
            x_cart = x_M16 + s*dx_dl + s * \
                np.mean(where2('dl', dl[i]))  # middle position
            # -- light path
            plt.plot([layout[t][0], layout[t][0], x_M16, x_cart+s*1, x_cart+s*1, x_IP, x_IP],
                     [layout[t][1], y_M12, y_M12, y_M12,
                         y_M12-wdl, y_M12-wdl, y_lab],
                     '-', alpha=0.5, linewidth=1.5, color=colors[i])
            # -- cart
            plt.plot(x_cart, y_M12-wdl/2, color=colors[i], markersize=22,
                     marker=DLm_W if (dl[i] == 3 or dl[i] == 4) else DLm_E)

            # -- write name of station where light is coming from
            if layout[t][1] > y_lab:  # north of the tracks
                plt.text(layout[t][0], y_lab-wdl, t, color=colors[i],
                         va='top', ha='center')
            else:  # south of the tracks
                plt.text(layout[t][0], fy_M12(7)+wdl, t, color=colors[i],
                         va='bottom', ha='center')
        # -- DL tunnel walls:
        U0 = layout['A0'][0] - 5
        U1 = layout['M0'][0] + 10
        color = (0.1, 0.2, 0.7)
        plt.fill_between([x_M16-110, x_M16+90], fy_M12(7)*np.array([1, 1]),
                         (fy_M12(7)-10)*np.array([1, 1]),
                         color=color, alpha=0.2, hatch='//')
        plt.fill_between([x_M16-110, x_M16+90], y_lab*np.array([1, 1]),
                         (y_lab+10)*np.array([1, 1]),
                         color=color, alpha=0.2, hatch='//')
        # end walls
        plt.fill_between((U0-10, U0), (fy_M12(7), fy_M12(7)), (y_lab, y_lab),
                         color=color, alpha=0.2, hatch='//')
        plt.fill_between((U1, U1+10), (fy_M12(7), fy_M12(7)), (y_lab, y_lab),
                         color=color, alpha=0.2, hatch='//')

        plt.ylim(fy_M12(7)-4*wdl, y_lab+4*wdl)
        plt.xlim(U0-4, U1+4)
        # plt.title('Delay line tunnel')
        ax2.get_yaxis().set_visible(False)
        ax2.get_xaxis().set_visible(False)

        # -- VCM pressure for each DL:
        # ax4 = plt.subplot(224, sharex=ax1)
        # for i,d in enumerate(dl):
        #     #plt.plot(res['lst'], res['vcm'][d], '-',label='VCM'+str(d))
        #     plt.plot(where('lst'), where2('vcm', d),
        #                 '-',markersize=8,label='VCM'+str(d),
        #                 alpha=0.7, color=colors[i], linewidth=3)
        #
        # plt.legend(loc='upper left', prop={'size':10}, numpoints=1,
        #            ncol=6)
        # plt.ylabel('DL VCM pressure %s(bar)'%('for AT/STS or UT' if STS else ''))
        # plt.xlabel('lst (h)')
        # ax1.set_xlim(res['lst'].min(), res['lst'].max())
        # plt.grid()
        # plt.ylim(0,5.0)
        # plt.hlines(2.5, res['lst'].min(), res['lst'].max(), color='k',
        #            alpha=0.4, linestyle='dashed', linewidth=3)

        print('tracking time:', res['lst'][np.where(res['observable'])[0][-1]] -
              res['lst'][np.where(res['observable'])[0][0]])
    else:
        return res


def projBaseline(T1, T2, target, lst, ip1=1, ip2=3, DL1=None, DL2=None,
                 min_altitude=30, max_OPD=100, max_airmass=None):
    """
    - T1, T2: telescopes stations (e.g. 'A0', 'U1', etc.)

    - target : [ra, dec] in decimal hour and decimal deg or name for SIMBAD

    - lst   : decimal hour (scalar or array)

    - ip1, ip2: input channels (optional)

    - DL1, DL2: delay lines (optional)

    min_altitude (default 20 degrees) sets the minimum altitude for
    observation. Checks also for shadowinf from the UTs.

    max_OPD (default 100m, from 11, to 111m) maximum OPD stroke of the delay lines.

    return a self explanatory dictionnary. units are m and degrees
    """
    if isinstance(target, str):
        # -- old simbad
        # s = simbad.query(target)[0]
        # radec = [s['RA.h'], s['DEC.d']]
        # --
        s = Simbad.query_object(target)
        ra = np.sum(np.float64(s['RA'][0].split())
                    * np.array([1, 1/60., 1/3600.]))
        dec = np.float64(s['DEC'][0].split())*np.array([1, 1/60., 1/3600.])
        dec = np.sum(np.abs(dec))*np.sign(dec[0])
        radec = ra, dec

    else:
        radec = target

    # -- hour angle
    ha = (np.array(lst) - radec[0]) * 360/24.0

    # -- alt-az
    dec = radec[1]

    tmp1 = np.sin(np.radians(dec)) * np.sin(np.radians(vlti_latitude)) +\
        np.cos(np.radians(dec)) * np.cos(np.radians(vlti_latitude)) *\
        np.cos(np.radians(ha))
    alt = np.arcsin(tmp1)*180/np.pi

    tmp1 = np.cos(np.radians(dec)) * np.sin(np.radians(ha))
    tmp2 = -np.cos(np.radians(vlti_latitude)) * np.sin(np.radians(dec)) + \
        np.sin(np.radians(vlti_latitude)) * np.cos(np.radians(dec)) * \
        np.cos(np.radians(ha))
    az = (360-np.arctan2(tmp1, tmp2)*180/np.pi) % 360

    b = [layout[T1][2]-layout[T2][2],
         layout[T1][3]-layout[T2][3],
         0.0]  # assumes you *DO NOT* combine ATs and UTs

    # -- projected baseline
    ch_ = np.cos(ha*np.pi/180.)
    sh_ = np.sin(ha*np.pi/180.)
    cl_ = np.cos(vlti_latitude*np.pi/180.)
    sl_ = np.sin(vlti_latitude*np.pi/180.)
    cd_ = np.cos(radec[1]*np.pi/180.)
    sd_ = np.sin(radec[1]*np.pi/180.)

    # -- (u,v) coordinates in m
    u = ch_ * b[0] - sl_*sh_ * b[1] + cl_*sh_ * b[2]
    v = sd_*sh_ * b[0] + (sl_*sd_*ch_+cl_*cd_) * b[1] -\
        (cl_*sd_*ch_ - sl_*cd_) * b[2]

    # -- static OPL to telescope
    if not DL1 is None and not DL2 is None:
        config = (T1+'DL%dIP%d' % (DL1, ip1), T2+'DL%dIP%d' % (DL2, ip2))
        # -- assume compressed, since it is the case for UT and AT-STS
        opl1, opl2 = computeOpl0(config, plot=False, compressed=True)
    else:
        opl1 = layout[T1][4] + 0.12*(ip1-1)
        opl2 = layout[T2][4] + 0.12*(ip2-1)

    # add telescope OPL (M3-M11 distance): only matters for UT-AT combination
    # email SMo 15/10/2013:
    # if T1.startswith('U'):
    #     opl1 += 65.80 # UT
    # else:
    #     opl1 += 18.19 # AT
    # if T2.startswith('U'):
    #     opl2 += 65.80 # UT
    # else:
    #     opl2 += 18.19 # AT

    # optical delay and optical path delay, in m
    d = -b[0]*cd_*sh_ -\
        b[1]*(sl_*cd_*ch_ - cl_*sd_) +\
        b[2]*(cl_*cd_*ch_ + sl_*sd_)
    d = np.array(d)

    # -- final OPD to correct for
    opd = d + opl2 - opl1

    # -- airmass:
    czt = np.cos(np.pi*(90-alt)/180.)
    airmass = 1/czt*(1-.0012*(1/czt**2-1))
    airmass *= (alt > 20.)
    airmass += (alt <= 20)*2.9

    # -- parallactic angle
    if np.cos(np.radians(dec)) != 0:
        parang = 180*np.arcsin(np.sin(np.radians(az)) *
                               np.cos(np.radians(vlti_latitude)) /
                               np.cos(np.radians(dec)))/np.pi
    else:
        parang = 0.

    # -- observability
    observable = (alt > min_altitude) * \
                 (np.abs(opd) < max_OPD) *\
                 (alt > np.interp(az % 360, horizon[T1][0], horizon[T1][1])) *\
                 (alt > np.interp(az % 360, horizon[T2][0], horizon[T2][1]))

    if not max_airmass is None:
        observable = observable*(airmass <= max_airmass)

    if isinstance(alt, np.ndarray):
        observable = np.array(observable)

    res = {'u': u, 'v': v, 'opd': opd, 'alt': alt, 'az': az,
           'observable': observable, 'parang': parang,
           'lst': lst, 'ra': radec[0], 'dec': radec[1],
           'B': np.sqrt(u**2+v**2), 'PA': np.arctan2(u, v)*180/np.pi,
           'airmass': airmass, 'baselines': T1+T2, 'opd': opd,
           'horizon': np.maximum(np.interp(az % 360, horizon[T1][0],
                                           horizon[T1][1]),
                                 np.interp(az % 360, horizon[T2][0],
                                           horizon[T2][1])),
           # -- baseline on the ground:
           'ground': np.array([layout[T2][0]-layout[T1][0],
                              layout[T2][1]-layout[T1][1],
                              (20. if 'U' in T2 else 0.)-(20. if 'U' in T1 else 0.)])}
    return res


def solveDLposition(dl, opd, dlPosMin=11/2., dlPosMax=111/2.,
                    stations=None, STS=False, constraints=None):
    """
    dl: list of delay lines numbers, like [1,3,2]
    opd: dict of OPDs for pairs of delay lines, in meters, for example,
    {(1,2):12.0, (3,1):-34.22, (3,2):0.0}. Convention is 2*(DL2-DL1)=12.0,
    2*(DL1-DL3)=-34.22, because the OPL is twice the DL position

    if a list of stations is given (same order as dl), VCM pressure will be
    computed.

    constraints: specific constraints for each DL, override the dlPosMin and
    dlPosMax which are global. e.g.: constraints={'min3':20, 'max5':40} means
    DL3 must be atleast at 20 m (position, i.e. 40m delay) and DL5 has to be
    at most at position 40 (i.e. delay 80)

    a bit smarter than the previous version: check len(dl) configuration,
    every time putting one of the DL closest to the lab.

    Author: amerand
    """
    # -- build default range constraints
    cons = {}
    for d in [1, 2, 3, 4, 5, 6]:
        cons['min'+str(d)] = dlPosMin
        cons['max'+str(d)] = dlPosMax
        if _test_dl34_plus30:
            # -- test the effect of extending DL3,4 for 30 OPL:
            if d in [3, 4]:
                cons['max'+str(d)] += 30

    if not constraints is None and isinstance(constraints, dict):
        for k in constraints.keys():
            if k in cons.keys():
                cons[k] = constraints[k]/2.
            else:
                print('WARNING: unknown constraint', k)

    # -- initialise dict used to track optimum solution:
    best = {}
    for d in dl:
        best[d] = np.nan
        best['vcm'+str(d)] = 5

    for i in range(len(dl)):  # -- for each DL
        # -- try to put the DL dl[i] at minimum position:
        dlPos = {dl[i]: cons['min'+str(dl[i])]}
        # -- explore all other DL positions
        while len(dlPos.keys()) != len(dl):
            for o in opd.keys():  # -- for each requested OPD
                # -- set DL pos according to OPD
                if o[0] in dlPos.keys() and not o[1] in dlPos.keys():
                    dlPos[o[1]] = dlPos[o[0]] - opd[o]/2.
                elif o[1] in dlPos.keys() and not o[0] in dlPos.keys():
                    dlPos[o[0]] = dlPos[o[1]] + opd[o]/2.
                # print(dlPos, dlPos.keys(), o[0], o[1])
        # -- check that DL carts are within optional constraints:
        test = all([dlPos[k] <= cons['max'+str(k)] for k in dlPos.keys()])
        test = all([dlPos[k] >= cons['min'+str(k)]
                   for k in dlPos.keys()]) and test

        # -- find best position
        if test:
            if not stations is None:
                # -- compute VCM pressure
                for k, d in enumerate(dl):
                    dlPos['vcm'+str(d)] = computeVcmPressure(stations[k],
                                                             d, 1, dlPos[d], STS=STS)
                # -- try to keep ALL VCM pressure to a minimum
                if np.max([dlPos['vcm'+str(d)] for d in dl]) <=\
                   np.max([best['vcm'+str(d)] for d in dl]):
                    best = dlPos.copy()
                # -- try to keep DL closer to the lab
                # if np.max([dlPos[d] for d in dl])<=np.max([best[d] for d in dl]):
                #    best = dlPos.copy()

            else:
                # -- set vcm pressure to nan
                for d in dl:
                    dlPos['vcm'+str(d)] = np.nan
                if np.mean([dlPos[d] for d in dl]) <= np.mean([best[d] for d in dl]):
                    best = dlPos.copy()
                # if np.max([dlPos[d] for d in dl]) <= np.max([best[d] for d in dl]):
                #    best = dlPos.copy()
        else:
            # print dlPos
            # print opd
            pass
    return best


_X, _Y = [], []


def visImage(image, scale, u, v, wl, debug=False):
    """
    image: 2D image, square pixels!
    scale: "scale", pixels size, in mas
    u, v: spatial coordinates (vectors, in m)
    wl: wavelength (vector or scalar, in um)
    """
    global _X, _Y
    if np.shape(_X) != np.shape(image) or \
            np.shape(_Y) != np.shape(image):
        Ny, Nx = np.shape(image)
        x, y = np.arange(Nx)-(Nx-1)/2, np.arange(Ny)-(Ny-1)/2
        _X, _Y = np.meshgrid(scale*x, scale*y)
        if debug:
            print('visImage: x', Nx, x.min(), x[Nx//2], x.max())
            print('visImage: y', Ny, y.min(), y[Ny//2], y.max())

    # -- fourier transform
    c = np.pi/180/3600/1000/1e-6
    if np.isscalar(wl):
        vis = image[:, :, None] *\
            np.exp(-2j*np.pi*c*(_X[:, :, None]*u[None, None, :] +
                                _Y[:, :, None]*v[None, None, :])/wl)
    else:
        vis = image[:, :, None, None] *\
            np.exp(-2j*np.pi*c*(_X[:, :, None, None]*u[None, None, :, None] +
                                _Y[:, :, None, None]*v[None, None, :, None]) /
                   wl[None, None, None, :])
    return np.sum(vis, axis=(0, 1))/np.sum(image)


def visCube(cube, u, v, wl):
    """
    cube:  {'image':, 'scale': 'WL':}
        image: Nwl (wavelength, optional) x Nx x Ny (spatial)
        scale: "scale", in mas
        wl: wavelength vector in um (ignored if cube is 2D)

    values for the computations:
        u, v: spatial coordinates (vectors, in m)
        wl: wavelength (vector, in um)
    """
    global _X, _Y
    res = np.zeros((len(u), len(wl)), np.complex64)
    _X, _Y = cube['X'], cube['Y']
    if False:
        # == Interpolate in visibility space =============================
        # -- compute V(u,v) for each wl of the cube:
        tmp = np.zeros((len(u), len(cube['WL'])), np.complex64)
        for i, x in enumerate(cube['WL']):
            tmp[:, i] = visImage(cube['image'][i, :, :], cube['scale'],
                                 u, v, cube['WL'][i])
        # -- interpolate
        for i in range(len(u)):
            res[i, :] = np.interp(wl, cube['WL'], tmp[i, :])
    else:
        # == Interpolate in image space ==================================
        for i, x in enumerate(wl):
            # -- find 2 closest images
            i1, i2 = np.argsort(np.abs(x-cube['WL']))[:2]
            # -- linear interpolation
            im = cube['image'][i1, :, :] + \
                (x-cube['WL'][i1])/(cube['WL'][i2]-cube['WL'][i1]) *\
                (cube['image'][i2, :, :]-cube['image'][i1, :, :])

            res[:, i] = visImage(im, cube['scale'], u, v, x)
    return res


def fluxCube(cube, wl):
    """
    cube:  {'image':, 'scale': 'WL':}
        image: Nx x Ny spatial pixels (x Nwl optional )
        scale: "scale", in mas (ignored)
        wl: wavelength vector in um (ignored if cube is 2D)
    wl: wavelength (vector, in um)
    """
    tmp = np.sum(cube['image'], axis=(1, 2))
    return np.interp(wl, cube['WL'], tmp)


def checkCubeShapes(cube, name='cube'):
    err = '\033[41mERROR\033[0m: '
    if type(cube) != dict:
        print(err+'"%s" should be dictionnary' % name)
        return False

    K = ['X', 'Y', 'image']
    if not all([k in cube for k in K]):
        print(err+'"%s" should at least contain "X", "Y" and "image"' % name)
        return False

    if type(cube['image']) != np.ndarray:
        print(err+'%s["image"] should be of type "np.ndarray"' % (name))
        return False

    if len(cube['image'].shape) == 2:
        L = {'X': 2, 'Y': 2, 'image': 2}
    else:
        L = {'WL': 1, 'X': 2, 'Y': 2, 'image': 3}

    if not all([type(cube[k]) == np.ndarray for k in L]):
        print(err+'%s should be of type "np.ndarray"' %
              (', '.join(['"%s"' % k for k in L])))
        for k in L:
            if type(cube[k]) != np.ndarray:
                print('  ', k, type(cube[k]), type(cube[k]) == np.ndarray)
        return False

    if not all([len(cube[k].shape) == L[k] for k in L]):
        print(err+'wrong shapes!')
        for k in L:
            if L[k] != len(cube[k].shape):
                print('  len(%s[%s].shape)=%d -> should be %d' % (name, k, L[k],
                                                                  len(cube[k].shape)))
        return False

    if cube['X'].shape != cube['Y'].shape:
        print(err+'"X" and "Y" should have same shapes!')
        print('  %s["X"].shape=%s != %s["Y"].shape=%s' % (name, str(cube['X'].shape),
                                                          name, str(cube['Y'].shape)))
        return False

    if L['image'] == 2:
        if cube['X'].shape != cube['image'].shape:
            print(err+'"image" has wrong dimensions!')
            print('  %s["X"].shape=%s != %s["image"].shape=%s' % (name, str(cube['X'].shape),
                                                                  name, str(cube['image'].shape)))
            return False
    else:
        _tup = tuple(list(cube['WL'].shape)+list(cube['X'].shape))
        if cube['image'].shape != _tup:
            print(err+'"image" has wrong dimensions!')
            print('  %s["image"].shape = %s, should be %s' % (name, str(cube['image'].shape),
                                                              str(_tup)))
            return False

    return True


__mjd0 = 57000


def makeFakeVLTI(t, target, lst, wl, mjd0=None, lst0=0,
                 diam=None, cube=None, noise=None, thres=None,
                 model=None, insname='synthetic', debug=False,
                 doubleDL=False):
    """
    for VLTI!

    t = list of stations
    targe = name of target, or (ra, dec) in (in hours, degs)
    lst = np.array of LST (in hours)
    wl = np.array of wavelength (in um)
    diam = UD diam, in mas. If set, returns simple UD model
    cube = {'image', 'scale', 'WL', 'spectrum'}
        image: Nx x Ny spatial pixels (x Nwl optional )
        scale: "scale", in mas (ignoired)
        wl: wavelength vector in um (ignored if cube is 2D only)
        spectrum: spectrum (same lemgth as wl), should be sum(image, axis=(0,1))
    model = a dictionnary describing a model (usual PMOIRED syntax)
    doubleDL = use double passage delay line (default=False)
    noise = noise to apply. If ==0, no noise; if None, typical noise (relative / degrees):
          {'V2': 0.01, '|V|':0.01, 'PHI':1., 'FLUX':0.01, 'T3PHI':1., 'T3AMP':0.01}
    """
    global __mjd0
    if noise == 0:
        noise = {k: 0 for k in ['V2', '|V|', 'PHI', 'FLUX', 'T3PHI', 'T3AMP']}
    if noise is None:
        # -- typical for GRAVITY good SNR
        noise = {'V2': 0.01, '|V|': 0.01, 'PHI': 0.5, 'FLUX': 0.01,
                 'T3PHI': 1., 'T3AMP': 0.01}
        # -- high precision
        # noise = {'V2': 0.001, '|V|':0.001, 'PHI':0.1, 'FLUX':0.001,
        #         'T3PHI':.1, 'T3AMP':0.001}
        #
    if type(noise)!=dict:
       raise Exception('noise should be a dictionnary such as '+
            str({'V2': 0.01, '|V|': 0.01, 'PHI': 0.5, 'FLUX': 0.01,'T3PHI': 1., 'T3AMP': 0.01}))

    # -- try to complete the noise dictionnary
    if 'V2' in noise and not '|V|' in noise:
        noise['|V|'] = noise['V2']
    if '|V|' in noise and not 'V2' in noise:
        noise['V2'] = noise['|V|']
    if '|V|' in noise and not 'T3AMP' in noise:
            noise['T3AMP'] = 1.5*noise['|V|']*1.5
    if 'T3PHI' in noise and not 'PHI' in noise:
        noise['PHI'] = noise['T3PHI']/1.5
    if 'PHI' in noise and not 'T3PHI' in noise:
            noise['T3PHI'] = noise['T3PHI']*1.5
    if not 'FLUX' in noise:
        noise['FLUX'] = 0.01

    if thres is None:
        # -- threshold for noise behavior (absolute)
        thres = {'V2': 0.01, '|V|': 0.02, 'T3AMP': 0.05, 'FLUX': 0.01}

    if doubleDL:
        tmp = nTelescopes(t, target, lst, max_OPD=200)
    else:
        tmp = nTelescopes(t, target, lst)
    if any(~tmp['observable']):
        print('WARNING: %2d out of %d LST are not observable!' %
              (np.sum(~tmp['observable']), len(lst)), end=' ')
        print('-'.join(t), np.array(lst)[~tmp['observable']])
        tmp['observable'] = np.array(tmp['observable'])
        for k in ['airmass', 'lst', 'alt', 'az', 'parang', 'horizon']:
            tmp[k] = np.array(tmp[k])[tmp['observable']]
        for k in ['u', 'v', 'B', 'PA', 'opd']:
            for b in tmp[k].keys():
                tmp[k][b] = tmp[k][b][tmp['observable']]
        lst = np.array(lst)[tmp['observable']]
        tmp['observable'] = tmp['observable'][tmp['observable']]
        # print(tmp)

    # -- fake MJD
    if mjd0 == None:
        # -- needed to simulate data taken at different dates
        mjd0 = __mjd0 + 1
        __mjd0 += 1
    tmp['MJD'] = (np.array(lst)-lst0)/24 + mjd0
    res = {'insname': '%s_%.3f_%.3fum_R%.0f' % (insname, min(wl), max(wl),
                                                np.mean(wl/np.gradient(wl))),
           'filename': 'synthetic',
           'targname': target if type(target) == str else
           '%.3f %.3f' % tuple(target),
           'pipeline': 'None',
           'header': {},
           'WL': wl,
           'telescopes': list(t),
           'baselines': tmp['baselines'],
           'TELLURICS': np.ones(len(wl)),
           'MJD': tmp['MJD'],
           }
    # print('res:', res)

    # == complex visibility function
    # -- default == unresolved gray object
    def fvis(u, v, l): return np.ones((len(u), len(l)))
    def fflux(l): return np.ones((len(lst), len(l)))

    if not cube is None:
        if not checkCubeShapes(cube):
            raise Exception('data cube is ill formed')
        if len(cube['image'].shape) == 2:
            def fvis(u, v, l): return visImage(
                cube['image'], cube['scale'], u, v, l)

            def fflux(l): return np.ones((len(lst), len(l)))
        else:
            def fvis(u, v, l): return visCube(cube, u, v, l)
            if 'spectrum' in cube:
                def fflux(l): return np.interp(l, cube['WL'], cube['spectrum'])
            else:
                def fflux(l): return np.interp(l, cube['WL'],
                                               np.sum(cube['image'], axis=(1, 2)))
    elif model is None and diam is None:
        # -- so we have at least somethinh to model!
        diam = 1.0

    if not diam is None:
        # -- simple uniform disk
        c = np.pi/180/3600/1000/1e-6

        def fvis(u, v, l):
            x = c*np.pi*diam*np.sqrt(u**2+v**2)[:, None]/wl[None, :]
            return 2*scipy.special.j1(x)/x

        def fflux(l): return np.ones((len(lst), len(l)))

    VIS = {}  # complex visibility
    # -- OI_VIS2
    OIV2 = {}
    # -- OI_VIS
    OIV = {}
    conf = {m: [] for m in tmp['MJD']}
    for b in tmp['baselines']:
        # -- complex visibility for this baseline
        VIS[b] = fvis(tmp['u'][b], tmp['v'][b], wl)
        nv2 = noise['V2']*np.maximum(np.abs(VIS[b])**2, thres['V2'])
        OIV2[b] = {'u': tmp['u'][b], 'v': tmp['v'][b],
                   'u/wl': tmp['u'][b][:, None]/wl[None, :],
                   'v/wl': tmp['v'][b][:, None]/wl[None, :],
                   'B/wl': tmp['B'][b][:, None]/wl[None, :],
                   # 'PA': tmp['PA'][b],
                   'PA': np.angle(tmp['v'][b][:, None]/wl[None, :] +
                                  1j*tmp['u'][b][:, None]/wl[None, :], deg=True),
                   'MJD': tmp['MJD'],
                   'MJD2': tmp['MJD'][:, None]+0*wl[None, :],
                   'FLAG': np.zeros((len(lst), len(wl)), bool),
                   'V2': np.abs(VIS[b])**2 +
                   nv2*np.random.randn(len(lst), len(wl)),
                   'EV2': nv2,
                   }
        if debug:
            print(b, list(OIV2[b].keys()))

        nv = noise['|V|']*np.maximum(np.abs(VIS[b]), thres['|V|'])
        OIV[b] = {'u': tmp['u'][b], 'v': tmp['v'][b],
                  'u/wl': tmp['u'][b][:, None]/wl[None, :],
                  'v/wl': tmp['v'][b][:, None]/wl[None, :],
                  'B/wl': tmp['B'][b][:, None]/wl[None, :],
                  # 'PA': tmp['PA'][b],
                  'PA': np.angle(tmp['v'][b][:, None]/wl[None, :] +
                                 1j*tmp['u'][b][:, None]/wl[None, :], deg=True),
                  'MJD': tmp['MJD'],
                  'MJD2': tmp['MJD'][:, None]+0*wl[None, :],
                  'FLAG': np.zeros((len(lst), len(wl)), bool),
                  '|V|': np.abs(VIS[b]) + nv*np.random.randn(len(lst), len(wl)),
                  'E|V|': nv,
                  'PHI': np.angle(VIS[b])*180/np.pi +
                  noise['PHI']*np.random.randn(len(lst), len(wl)),
                  'EPHI': noise['PHI']*np.ones((len(lst), len(wl))),
                  }
        for m in tmp['MJD']:
            conf[m].append(b)
    res['OI_VIS2'] = OIV2
    res['OI_VIS'] = OIV

    # -- OI_FLUX
    OIF = {}
    for s in t:
        OIF[s] = {'FLUX': fflux(wl)*(1 +
                                     noise['FLUX']*np.random.randn(len(lst), len(wl))),
                  'RFLUX': fflux(wl)*(1 +
                                      noise['FLUX']*np.random.randn(len(lst), len(wl))),
                  'EFLUX': noise['FLUX']*np.ones((len(lst), len(wl)))*fflux(wl)[None, :],
                  'FLAG': np.zeros((len(lst), len(wl)), bool),
                  'MJD': tmp['MJD'],
                  'MJD2': tmp['MJD'][:, None]+0*wl[None, :],
                  }
        for m in tmp['MJD']:
            conf[m].append(s)
    res['OI_FLUX'] = OIF

    # -- OI_T3
    OIT3 = {}
    res['triangles'] = []
    if len(t) >= 3:
        # print(tmp['baselines'])
        for i, tri in enumerate(itertools.combinations(t, 3)):
            res['triangles'].append(''.join(tri))
            if tri[0]+tri[1] in tmp['baselines']:
                b1 = tri[0]+tri[1]
                s1 = 1
                T3 = VIS[b1].copy()
            elif tri[1]+tri[0] in tmp['baselines']:
                b1 = tri[1]+tri[0]
                s1 = -1
                T3 = np.conj(VIS[b1])

            if tri[1]+tri[2] in tmp['baselines']:
                b2 = tri[1]+tri[2]
                s2 = 1
                T3 *= VIS[b2]
            elif tri[2]+tri[1] in tmp['baselines']:
                b2 = tri[2]+tri[1]
                s2 = -1
                T3 *= np.conj(VIS[b2])

            if tri[2]+tri[0] in tmp['baselines']:
                b3 = tri[2]+tri[0]
                s3 = 1
                T3 *= VIS[b3]
            elif tri[0]+tri[2] in tmp['baselines']:
                b3 = tri[0]+tri[2]
                s3 = -1
                T3 *= np.conj(VIS[b3])

            # -- minimum visibility
            minV = np.minimum(np.minimum(np.abs(VIS[b1]), np.abs(VIS[b2])),
                              np.abs(VIS[b3]))

            form = ('+' if s1 > 0 else '-')+b1 +\
                   ('+' if s2 > 0 else '-')+b2 +\
                   ('+' if s3 > 0 else '-')+b3
            # print('debug:', i, tri, form)

            # -- KLUDGE -> account for T3 amplitude
            ncp = 1  # np.maximum(1, (thres['T3AMP']/np.abs(T3))**(1/3))
            # ncp = np.maximum(1, thres['T3AMP']/minV)
            if debug:
                print(tri, 'NCP:', ncp.min(), ncp.max(),
                      'noise T3PHI:', noise['T3PHI'])

            OIT3[''.join(tri)] = {
                'MJD': tmp['MJD'],
                'MJD2': tmp['MJD'][:, None]+0*wl[None, :],
                'FLAG': np.zeros((len(lst), len(wl)), bool),
                'T3AMP': np.abs(T3)*(1 +
                                     noise['T3AMP']*np.random.randn(len(lst), len(wl))),
                'ET3AMP': noise['T3AMP']*np.ones((len(lst), len(wl)))*np.abs(T3),
                'T3PHI': np.angle(T3)*180/np.pi +
                ncp*noise['T3PHI']*np.random.randn(len(lst), len(wl)),
                'ET3PHI': ncp*noise['T3PHI']*np.ones((len(lst), len(wl))),
                'u1': s1*tmp['u'][b1], 'v1': s1*tmp['v'][b1],
                'u2': s2*tmp['u'][b2], 'v2': s2*tmp['v'][b2],
                'u1/wl': s1*tmp['u'][b1][:, None]/wl[None, :],
                'v1/wl': s1*tmp['v'][b1][:, None]/wl[None, :],
                'u2/wl': s2*tmp['u'][b2][:, None]/wl[None, :],
                'v2/wl': s2*tmp['v'][b2][:, None]/wl[None, :],
                'B1': tmp['B'][b1],
                'B2': tmp['B'][b2],
                'B3': tmp['B'][b3],
                'Bmin/wl': np.minimum(tmp['B'][b1], tmp['B'][b2], tmp['B'][b3])[:, None] /
                wl[None, :],
                'Bmax/wl': np.maximum(tmp['B'][b1], tmp['B'][b2], tmp['B'][b3])[:, None] /
                wl[None, :],
                'Bavg/wl': (tmp['B'][b1]+tmp['B'][b2]+tmp['B'][b3])[:, None] /
                wl[None, :]/3,
                'formula': [[s1, s2, s3], [b1, b2, b3],
                            list(range(len(tmp['MJD']))),
                            list(range(len(tmp['MJD']))),
                            list(range(len(tmp['MJD']))),]
            }
            # print(''.join(tri), sorted(OIT3[''.join(tri)].keys()))
            for m in tmp['MJD']:
                conf[m].append(''.join(tri))
    res['OI_T3'] = OIT3

    if not model is None:
        param = model
        res['fit'] = {'obs': ['V2', '|V|', 'T3PHI', 'T3AMP', 'PHI', 'FLUX']}
        res = oimodels.VmodelOI(res, param, fullOutput=True)

        # --
        for k in ['PA']:
            for b in res['OI_VIS']:
                res['OI_VIS'][b][k] = tmp[k][b][:, None]+0*res['WL'][None, :]
            for b in res['OI_VIS2']:
                res['OI_VIS2'][b][k] = tmp[k][b][:, None]+0*res['WL'][None, :]

        # -- add noise
        def addnoise(e, k, o):
            res[e][k][o] += np.random.randn(res[e][k][o].shape[0],
                                            res[e][k][o].shape[1]) *\
                res[e][k]['E'+o]

        if 'OI_FLUX' in res:
            for k in res['OI_FLUX'].keys():
                res['OI_FLUX'][k]['EFLUX'] = noise['FLUX']*np.maximum(res['OI_FLUX'][k]['FLUX'],
                                                                      thres['FLUX'])
                addnoise('OI_FLUX', k, 'FLUX')

        for k in res['OI_VIS'].keys():
            res['OI_VIS'][k]['E|V|'] = noise['|V|'] * \
                np.maximum(res['OI_VIS'][k]['|V|'], thres['|V|'])
            # res['OI_VIS'][k]['EPHI'] = noise['PHI']*(thres['|V|']/np.sqrt(thres['|V|']**2 +
            #                                np.maximum(res['OI_VIS'][k]['|V|'], thres['|V|'])**2))
            res['OI_VIS'][k]['EPHI'] = noise['PHI'] * \
                np.ones(res['OI_VIS'][k]['PHI'].shape)
            res['OI_VIS2'][k]['EV2'] = noise['V2'] * \
                np.maximum(res['OI_VIS2'][k]['V2'], thres['V2'])
            addnoise('OI_VIS', k, '|V|')
            addnoise('OI_VIS', k, 'PHI')
            addnoise('OI_VIS2', k, 'V2')

        for k in res['OI_T3'].keys():
            res['OI_T3'][k]['ET3AMP'] = noise['T3AMP'] * \
                np.maximum(res['OI_T3'][k]['T3AMP'], thres['T3AMP'])
            # res['OI_T3'][k]['ET3PHI'] = noise['T3PHI']*(thres['T3AMP']/np.sqrt(thres['T3AMP']**2+
            #                     np.maximum(res['OI_T3'][k]['T3AMP'], thres['T3AMP'])**2))
            res['OI_T3'][k]['ET3PHI'] = noise['T3PHI'] * \
                np.ones(res['OI_T3'][k]['T3PHI'].shape)

            addnoise('OI_T3', k, 'T3AMP')
            addnoise('OI_T3', k, 'T3PHI')

    res['configurations per MJD'] = conf
    return res


makeFake = makeFakeVLTI  # legacy



# oicandid.py

import numpy as np
import matplotlib.pyplot as plt
import time
import scipy

from matplotlib.colors import LinearSegmentedColormap

import multiprocessing

from pmoired import oimodels

def fitMap(oi, firstGuess=None, companion='c', fitAlso=[],
            rmin=None, rmax=None, rstep=None,
            multi=True, fig=1, cmap=None, doNotFit=[], logchi2=False):
    """
    firstguess: model. Star should be parametrized with '*,...' and companion 'c,...'.
        Other components can be used as well in the initial guess which can be
        passed in the variable "firstguess". Fit Only
    """
    D = {'|V|':('OI_VIS', 'B/wl'),
         'V2':('OI_VIS2', 'B/wl'),
         'T3PHI':('OI_T3', 'Bmax/wl'),
        }
    if type(oi) == dict:
        oi = [oi]
    for o in oi:
        assert 'fit' in o, "'fit' should be defined!"
        # -- in m/um
        _bmaxR, _bmax = 0., 0.
        for e in o['fit']['obs']:
            R = np.mean(o['WL'])/np.diff(o['WL']).mean()
            if e in D:
                for b in o[D[e][0]].keys():
                    #print(o[D[e][0]][b].keys())
                    _bmax = max(_bmax, np.max(o[D[e][0]][b][D[e][1]]))
                    _bmaxR = max(_bmaxR, np.max(o[D[e][0]][b][D[e][1]])/R)

    print('max(B/wl), max(B/wl/R):', _bmax, _bmaxR)
    if rstep is None:
        rstep = 206.24/_bmax # one fringe resolution
        print('> rstep set to %.2f mas based on largest (B/wl)'%rstep)
    if rmin is None:
        rmin = rstep
        print('> rmin set to %.2f mas based on rstep'%rmin)
    if rmax is None:
        rmax = 40
        rmax = 0.5*206.265/_bmaxR # 1/2 full smearing
        print('> rmax set to %.1f mas based on largest (B/wl/R)'%rmax)

    if cmap is None:
        colors = [(0,1,1), 'yellow', 'orange', 'darkorange', 'red', 'darkred', 'black']
        nodes = [0.0, 0.2,0.4, 0.6, 0.8, 0.9, 1]
        nodes = np.linspace(0,1,len(colors))**1.5
        #colors = [(0,1,1), 'blue', 'black', 'darkred', 'red', 'darkorange', 'orange', 'yellow']
        #nodes = [0.0,  0.05, 0.1, 0.3, 0.5, 0.7, 0.9, 1]
        cmap = LinearSegmentedColormap.from_list("candid", list(zip(nodes, colors)))

    # -- single star
    if type(firstGuess)==dict and '*,ud' in firstGuess:
        UD = oimodels.fitOI(oi, {'ud':firstGuess['*,ud']}, verbose=False)
    elif type(firstGuess)==dict and '*,diam' in firstGuess:
        UD = oimodels.fitOI(oi, {'ud':firstGuess['*,diam']}, verbose=False)
    else:
        UD = oimodels.fitOI(oi, {'ud':1.0}, verbose=False)
    print('uniform disk fit: chi2=%.3f'%UD['chi2'])
    default = {'*,ud':0.0, '*,f':1.0, companion+',ud':0.0, companion+',f':0.01}

    if firstGuess is None:
        firstGuess = default
    
    default.update(firstGuess)
    firstGuess = default

    if '*,diam' in firstGuess and '*,ud' in firstGuess:
        firstGuess.pop('*,ud')
    print('firstGuess:', firstGuess)


    fitOnly = ['c,f', 'c,x', 'c,y']
    if type(doNotFit) == list:
        for k in doNotFit:
            if k in fitOnly:
                fitOnly.remove(k)

    if type(fitAlso) == list:
        fitOnly.extend(fitAlso)

    kwargs = {'maxfev':10000, 'ftol':1e-5, 'verbose':False,
              'fitOnly':fitOnly, }

    X, Y = [], []
    for r in np.linspace(rmin, rmax, int((rmax-rmin)/rstep+1)):
        t = np.linspace(0, 2*np.pi, max(int(2*np.pi*r/rstep+1), 5))[:-1]
        X.extend(list(r*np.cos(t)))
        Y.extend(list(r*np.sin(t)))

    if type(fig) == int:
        plt.close(fig)
        plt.figure(fig, figsize=(9, 4))
        ax1 = plt.subplot(1,2,1, aspect='equal')
        ax2 = plt.subplot(2,2,2)
        ax3 = plt.subplot(2,2,4)
        # -- starting points on grid
        ax1.plot(X, Y, '+k', alpha=0.1)
        # -- central star
        ax1.plot(0, 0, '*c', label='central star')
        N = len(X)
        fig = True
    else:
        fig = False

    res = []
    if multi:
        if type(multi)!=int:
            Np = min(multiprocessing.cpu_count(), N)
        else:
            Np = min(multi, N)
        print('running', N, 'fits...')
        # -- estimate fitting time by running 'Np' fit in parallel
        pool = multiprocessing.Pool(Np)
        t = time.time()
        for i in range(min(Np, N)):
            tmp = firstGuess.copy()
            tmp['c,x'] = X[i]
            tmp['c,y'] = Y[i]
            #print(tmp)
            res.append(pool.apply_async(oimodels.fitOI, (oi, tmp, ), kwargs))
        pool.close()
        pool.join()
        print('initial estimate: %.1f fit per minute using %d threads'%(3600/(time.time()-t), Np))
        res = [r.get(timeout=60) for r in res]
        # -- run the remaining
        if N>Np:
            pool = multiprocessing.Pool(Np)
            for i in range(max(N-Np, 0)):
                tmp = firstGuess.copy()
                tmp['c,x'] = X[Np+i]
                tmp['c,y'] = Y[Np+i]
                res.append(pool.apply_async(oimodels.fitOI, (oi, tmp, ), kwargs))
            pool.close()
            pool.join()
            res = res[:Np]+[r.get(timeout=1) for r in res[Np:]]
    else:
        Np = 1
        tmp = firstGuess.copy()
        tmp['c,x'] = X[0]
        tmp['c,y'] = Y[0]
        t = time.time()
        res.append(oimodels.fitOI(oi, tmp, **kwargs))
        print('one fit takes ~%.2fs'%(time.time()-t))
        for i in range(N-1):
            tmp = firstGuess.copy()
            tmp['c,x'] = X[Np+i]
            tmp['c,y'] = Y[Np+i]
            res.append(oimodels.fitOI(oi, tmp, **kwargs))
    print('it took %.1fs, %.0f fit per minute on average (%.0fms per fit)'%(
            time.time()-t, 60/(time.time()-t)*N, 1000*(time.time()-t)/N))
    ncall = np.sum([r['info']['nfev'] for r in res])
    print('    %d calls, %.1f calls per fit, %.2fms per call'%(ncall,
            ncall/N, 1000*(time.time()-t)/ncall))

    # -- fit converged
    _n = len(res)
    res = [r for r in res if np.sum([r['uncer'][k] for k in r['uncer']])]
    if len(res)<_n:
        print('  - removing %d fits which did not converge properly'%(_n-len(res)))

    # -- solution within explored range
    _n = len(res)
    res = [r for r in res if
            np.sqrt(r['best']['c,x']**2+r['best']['c,y']**2)>=rmin and
            np.sqrt(r['best']['c,x']**2+r['best']['c,y']**2)<=rmax]
    if len(res)<_n:
        print('  - removing %d fits outside search range'%(_n-len(res)))

    # -- chi2 cannot be larger than 1.1*chi2 single UD
    _n = len(res)
    res = [r for r in res if r['chi2']<1.1*UD['chi2']]
    if len(res)<_n:
        print('  - removing %d fits chi2 > 1.1*chi2_single_star'%(_n-len(res)))

    n=0
    for r in res:
        if r['best']['c,f']>r['best']['*,f']:
            n +=1
            r['best']['c,x'] *= -1
            r['best']['c,y'] *= -1
            r['best']['c,f'] *= 1/r['best']['c,f']
    if n>0:
        print('  %d fits corrected for c,f>*,f'%n)

    # -- show displacement of fit:
    for r in res:
        r['displ'] = np.sqrt((r['firstGuess']['c,x']-r['best']['c,x'])**2+
                            (r['firstGuess']['c,y']-r['best']['c,y'])**2)
        if fig:
            ax1.plot([r['firstGuess']['c,x'], r['best']['c,x']],
                     [r['firstGuess']['c,y'], r['best']['c,y']],
                     '-k', alpha=0.1)

    # -- filter to unique solutions:
    uni = [res[0]]
    for r in res[1:]:
        dupl = False
        if not dupl:
            for u in uni:
                d = 0
                for k in u['fitOnly']:
                    d += (u['best'][k] - r['best'][k])**2/\
                         (u['uncer'][k]**2 + r['uncer'][k]**2)
                if d<len(u['fitOnly']):
                    dupl = True
                    continue
        if not dupl:
            uni.append(r)
    print('%d unique minima from %d fits'%(len(uni), len(res)))
    uni = sorted(uni, key=lambda x: x['chi2'])
    print('best fit chi2 = %.3f'%uni[0]['chi2'])
    print('{')
    for k in uni[0]['best'].keys():
        if uni[0]['uncer'][k]>0:
            n = max(int(2-np.log10(uni[0]['uncer'][k])), 0)
            f = '%.'+str(n)+'f'
            f = f+', # +- '+f
            print('  ',"'"+k+"':", f%(uni[0]['best'][k], uni[0]['uncer'][k]))
        else:
            print('  ',"'"+k+"':", uni[0]['best'][k], ',')
    print('}')
    r = np.sqrt(uni[0]['best']['c,x']**2+uni[0]['best']['c,y']**2)
    print('separation: %.4f mas'%r)
    pa = np.arctan2(uni[0]['best']['c,x'], uni[0]['best']['c,y'])
    pa *= 180/np.pi
    print('PA: %.2f degress'%pa)

    if fig:
        # -- best solution:
        #ax1.plot(uni[0]['best']['c,x'], uni[0]['best']['c,y'], '1r')
        #ax1.plot(uni[0]['best']['c,x'], uni[0]['best']['c,y'], '2r')
        #ax1.plot(uni[0]['best']['c,x'], uni[0]['best']['c,y'], '*y',
        #        label='best guess companion')
        c = [f['chi2'] for f in uni]
        if logchi2:
            c = np.log10(c)
        sca = ax1.scatter([f['best']['c,x'] for f in uni],
                          [f['best']['c,y'] for f in uni],
                         c = c,
                    cmap=cmap)
        # -- check that companion flux is less than central star flux
        if logchi2:
            plt.colorbar(sca, label=r'log$_{10}\chi^2$', ax=ax1)
        else:
            plt.colorbar(sca, label=r'$\chi^2$', ax=ax1)
        ax1.invert_xaxis()
        ax1.set_title('valid local minima')
        ax1.set_xlabel(r'x = RA offset (E$\leftarrow$ , mas)')
        ax1.set_ylabel(r'y = dec offset ($\rightarrow$N, mas)')

        X = [r['displ'] for r in res]
        bins = int(5*np.ptp(X)/np.std(X))

        ax2.hist(X, bins=bins, align='mid', color='k', histtype='step', alpha=0.9)
        ax2.hist(X, bins=bins, align='mid', color='k', histtype='stepfilled', alpha=0.05)
        ax2.vlines(rstep, 0, ax2.get_ylim()[1], label='grid step')
        ax2.set_xlabel('fit displacement (mas)')
        ax2.legend()

        X = [r['chi2'] for r in res]
        if logchi2:
            X = np.log10(X)
        bins = int(5*np.ptp(X)/np.std(X))
        ax3.hist(X, bins=bins, align='mid', color='k', histtype='step', alpha=0.9)
        ax3.hist(X, bins=bins, align='mid', color='k', histtype='stepfilled', alpha=0.05)
        if logchi2:
            ax3.vlines(np.log10(UD['chi2']), 0, ax3.get_ylim()[1], label='single star fit', color='r')
            ax3.set_xlabel('local minima log$_{10}\chi^2$')
        else:
            ax3.vlines(UD['chi2'], 0, ax3.get_ylim()[1], label='single star fit', color='r')
            ax3.set_xlabel('local minima $\chi^2$')

        ax3.legend(loc='upper left')
        plt.tight_layout()
    return uni

def _nSigmas(chi2r_TEST, chi2r_TRUE, NDOF):
    """
    - chi2r_TEST is the hypothesis we test
    - chi2r_TRUE is what we think is what described best the data
    - NDOF: numer of degres of freedom

    chi2r_TRUE <= chi2r_TEST

    returns the nSigma detection
    """
    p = scipy.stats.chi2.cdf(NDOF, NDOF*chi2r_TEST/chi2r_TRUE)
    log10p = np.log10(np.maximum(p, 1e-161)) ### Alex: 50 sigmas max
    #p = np.maximum(p, -100)
    res = np.sqrt(scipy.stats.chi2.ppf(1-p,1))
    # x = np.logspace(-15,-12,100)
    # c = np.polyfit(np.log10(x), np.sqrt(scipy.stats.chi2.ppf(1-x,1)), 1)
    c = np.array([-0.29842513,  3.55829518])
    if isinstance(res, np.ndarray):
        res[log10p<-15] = np.polyval(c, log10p[log10p<-15])
        res = np.nan_to_num(res)
        res += 90*(res==0)
    else:
        if log10p<-15:
            res =  np.polyval(c, log10p)
        if np.isnan(res):
            res = 90.
    return res



import numpy as np
"""
Daubechies Wavelet, D4 is the default behaviour, but other orders are available.

http://en.wikipedia.org/wiki/Daubechies_wavelet
http://www.bearcave.com/misl/misl_tech/wavelets/daubechies/index.html
http://cam.mathlab.stthomas.edu/wavelets/pdffiles/UST06/Lecture6.pdf
"""

coefH = {}
coefG = {}

# http://en.wikipedia.org/wiki/Daubechies_wavelet
# coefficients have been double checked
coefH[2]=np.array([1.0,1.0])
coefH[4]=np.array([0.6830127,1.1830127,0.3169873,-0.1830127])
coefH[6]=np.array([0.47046721,1.14111692,0.650365,-0.19093442,-0.12083221,0.0498175])
coefH[8]=np.array([0.32580343,1.01094572,0.8922014,-0.03957503,
                  -0.26450717,0.0436163,0.0465036,-0.01498699])
coefH[10]=np.array([0.22641898,0.85394354,1.02432694,0.19576696,-0.34265671,
                    -0.04560113,0.10970265,-0.00882680,-0.01779187,4.71742793e-3])
coefH[12]=np.array([0.15774243,0.69950381,1.06226376,0.44583132,
                    -0.31998660,-0.18351806,0.13788809,0.03892321,
                    -0.04466375,7.83251152e-4,6.75606236e-3,-1.52353381e-3])
coefH[14]=np.array([0.11009943,0.56079128,1.03114849,0.66437248,
                    -0.20351382,-0.31683501,0.1008467,0.11400345,
                    -0.05378245,-0.02343994,0.01774979,
                    6.07514995e-4,-2.54790472e-3,5.00226853e-4])
coefH[16]=np.array([0.07695562,0.44246725,0.95548615,0.82781653,
                    -0.02238574,-0.40165863,6.68194092e-4,0.18207636,
                    -0.02456390,-0.06235021,0.01977216,0.01236884,
                    -6.88771926e-3,-5.54004549e-4,9.55229711e-4,-1.66137261e-4])
coefH[18]=np.array([0.05385035,0.34483430,0.85534906,0.92954571,
                    0.18836955,-0.41475176,-0.13695355,0.21006834,
                    0.043452675,-0.09564726,3.54892813e-4,0.03162417,
                    -6.67962023e-3,-6.05496058e-3,2.61296728e-3,
                    3.25814671e-4,-3.56329759e-4,5.5645514e-5])
coefH[20]=np.array([0.03771716,0.26612218,0.74557507,0.97362811,
                    0.39763774,-0.35333620,-0.27710988,0.18012745,
                    0.13160299,-0.10096657,-0.04165925,0.04696981,
                    5.10043697e-3,-0.01517900,1.97332536e-3,2.81768659e-3,
                    -9.69947840e-4,-1.64709006e-4,1.32354367e-4,-1.875841e-5])

for k in coefH.keys():
    # -- direct transform
    coefH[k] /= np.sqrt(2)
    # -- inverse transform
    coefG[k] = (-1.)**np.arange(k)*coefH[k][::-1]

# import scipy.signal
# for i in range(34):
#     k = 2*i+2
#     coefH[k] = scipy.signal.daub(i+1)
#     coefG[k] = (-1.)**np.arange(k)*coefH[k][::-1]

fMats = {} # stores forward transform matrices
iMats = {} # stores inverse transform matrices

def test():
    import scipy.misc
    from matplotlib import pyplot as plt

    plt.close(1)
    plt.figure(1)
    plt.clf()
    # -- 1D signal
    y = np.sin(np.arange(128)/4.)+0.1*np.random.randn(128)
    plt.subplot(121)
    plt.title('signal')
    plt.plot(y)
    plt.subplot(122)
    plt.title('transform')
    for o in range(4):
        plt.plot(oneD(y, 5, order=2*(o+1)), label='O=%d'%(2*(o+1)))
    plt.legend()

    plt.close(2)
    plt.figure(2, figsize=(12,6))
    plt.clf()

    # -- test image
    image = scipy.misc.ascent()
    #image = scipy.misc.face()

    plt.subplot(121, frameon=False)
    plt.imshow(image, cmap='bone')
    plt.subplot(122, frameon=False)
    plt.imshow(np.abs(twoD(image, 1, order=2))**0.5, cmap='gist_stern')
    return

def forwardMat(n, p=4, debug=False):
    """
    compute matrix
    """
    assert (p in coefH.keys()), "coef for order "+str(p)+" are not implemented"
    if (n,p) in fMats.keys() and not debug:
        return fMats[(n,p)]
    mat = np.zeros((n,n))
    deb = [['--' for k in range(n)] for l in range(n)]
    deb = np.array(deb)
    for k in range(n)[::2]:
        for i in range(p):
            mat[k,(k+i)%n]   += coefH[p][i]
            mat[k+1,(k+i)%n] += coefG[p][i]
            deb[k,(k+i)%n]   = 'H'+str(i)
            deb[k+1,(k+i)%n] = 'G'+str(i)
    mat = np.roll(mat, 0, axis=0)
    deb = np.roll(deb, 0, axis=0)

    fMats[(n,p)] = mat
    iMats[(n,p)] = np.linalg.inv(mat)
    if debug:
        return deb
    return mat

def inverseMat(n,p=4, debug=False):
    if not (n,p) in iMats.keys():
        forwardMat(n,p, debug)
    return iMats[(n,p)]

#@numba.jit
def oneD(x,n,order=4):
    """
    1D D4 transform of x, ran n times recursively. If n<0, then the inverse
    transform is applied.

    len(x) should be a power of 2
    """
    res = np.array(x).copy()
    p = len(x)//2
    if n>0: # -- forward transform
        tmp = np.dot(forwardMat(len(x),order), x)
        res[p:] = tmp[1::2] # -- wavelet
        if n>1: # -- recursion
            res[:p] = oneD(tmp[0::2], n-1, order=order)/np.sqrt(2)
        else:
            res[:p] = tmp[0::2] # -- scaling
    elif n<0: # -- inverse transform
        tmp = np.array(x).copy()
        if n<-1: # -- recursion
            tmp[:p] = oneD(res[:p], n+1, order=order)
        res[0::2] = tmp[:p]*np.sqrt(2) # -- scaling
        res[1::2] = tmp[p:] # -- wavelet
        res = np.dot(inverseMat(len(x), order), res)
    return res

def build1DFilt(N,filt):
    """
    N is the size of the signal, filt is the filter. First value is the
    transmission for the lowest freq, the last value is for the highest
    frequency.

    for example:
    filt=[1,0.5,0] is a low pass filter
    filt=[0,0.2,0.4,0.8,1] is a high pass filter

    """
    res = np.zeros(N)
    p = N//2
    res[p:]=filt[-1]
    if len(filt)>2:
        res[:p]=build1DFilt(p,filt[:-1])
    else:
        res[:p]=filt[0]
    return res

#@numba.jit
def filter1D(x,filt,order=4):
    """
    x = 1D signal, filt is a transmission vector.

    x can have any dims (will be padded if needed)

    filt can be, for example:
    [1,0.5,0] for a low pass filter
    [0,0.2,0.4,0.8,1] for a high pass filter
    """
    N = np.log(len(x))/np.log(2)
    if N>int(N):
        N=int(N+1)
    else:
        N=int(N)
    tmp = np.zeros(2**N)
    P = 2**N-len(x)
    # -- zero padding
    #tmp[:len(x)]=x
    # -- mirror padding
    if P>0:
        tmp[P//2:P//2+len(x)]=x
        tmp[:P//2+1] = x[P//2::-1]
        tmp[-P//2:] = x[-2:-P//2-2:-1]
    else:
        tmp = x

    res = oneD(oneD(tmp,len(filt)+1,order)*
                build1DFilt(2**N,filt),
                -(len(filt)+1),order)
    # -- zero padding:
    if P>0:
        return res[P//2:P//2+len(x)]
    else:
        return res

def twoD(x,n,order=4):
    """
    2D D4 transform of x, ran n times recursively. If n<0, then the inverse
    transform is applied.

    only for x a square array, of dims being a power of 2
    """
    res = np.array(x).copy()
    p = x.shape[0]//2
    if n>0:
        for k in range(x.shape[0]):
            res[:,k] = oneD(res[:,k],1,order)
        for k in range(x.shape[0]):
            res[k,:] = oneD(res[k,:],1,order)
        if n>1:
            res[:p,:p] = twoD(res[:p,:p],n-1,order)
    elif n<0:
        if n<-1:
            res[:p,:p] = twoD(res[:p,:p],n+1,order)
        for k in range(x.shape[0]):
            res[k,:]=oneD(res[k,:],-1,order)
        for k in range(x.shape[0]):
            res[:,k]=oneD(res[:,k],-1,order)
    return res

def build2DFilt(N,filt):
    """
    NxN is the size of the image, filt is the filter. First value is the
    transmission for the lowest freq, the last value is for the highest
    frequency.

    for example:
    filt=[1,0.5,0] is a low pass filter
    filt=[0,0.2,0.4,0.8,1] is a high pass filter

    """
    res = np.zeros((N,N))
    p = N//2
    res[p:,p:]=filt[-1]
    res[:p,p:]=filt[-1]
    res[p:,:p]=filt[-1]
    if len(filt)>2:
        res[:p,:p]=build2DFilt(p,filt[:-1])
    else:
        res[:p,:p]=filt[0]
    return res

def filter2D(x,filt,order=4):
    """
    x = 2D image, filt is a transmission vector.

    x can have any dims (will be padded if needed)

    filt can be, for example:
    [1,0.5,0] for a low pass filter
    [0,0.2,0.4,0.8,1] for a high pass filter
    """
    N = int(np.log(max(x.shape))/np.log(2))+1
    tmp = np.zeros((2**N,2**N))
    tmp[:x.shape[0],:x.shape[1]]=x
    return twoD(twoD(tmp,len(filt)+1,order)*
                build2DFilt(tmp.shape[0],filt),
                -(len(filt)+1),order)[:x.shape[0],:x.shape[1]]

def struct2D(x,order=4):
    """
    structured 2D transform. x must have (2**N, 2**N) shape.

    returns a dictionnary with separate frequencies and indexes for
    reconstruction. fx_X where x s a number from 1...N, 1 being the lowest
    frequency, X=D, H or V for diagonal, horizontal or vertical frequencies.f0
    are the lowest frequencies (residuals).

    keys in the dictionnary:
    N = log2(width)
    T = unstructured transform
    frequencies for i=0 to i=N-1:
        f0: residuals (2x2)
        f1_D, f1_H, f1_V: diagonal, horizontal and vertical frequencies (2x2 each)
        ...
        fi_D, fi_H, fi_V: diagonal, horizontal and vertical frequencies (2**ix2**i each)
    """
    N = int(np.log(min(x.shape))/np.log(2))
    t = twoD(x, N, order)
    res = {'N':N, 'T':t}
    n = 0
    for k in range(N):
        if k==0: # -- residuals
            #res['f'+str(k)] = t[n:n+2**k, n:n+2**k] # -- amplitudes
            #res['i'+str(k)] = (n,n+2**k, n,n+2**k) # -- indices
            res['f'+str(k)] = t[0:2, 0:2] # -- amplitudes
            res['i'+str(k)] = (0,2, 0,2) # -- indices
            n += 2
        else: # -- frequencies: Diagonal, Horizontal and Vertical
            res['f'+str(k)+'_D'] = t[n:n+2**k, n:n+2**k]
            res['i'+str(k)+'_D'] =  (n,n+2**k, n,n+2**k)
            res['f'+str(k)+'_H'] = t[n:n+2**k, 0:2**k]
            res['i'+str(k)+'_H'] =  (n,n+2**k, 0,2**k)
            res['f'+str(k)+'_V'] = t[0:2**k, n:n+2**k]
            res['i'+str(k)+'_V'] =  (0,2**k, n,n+2**k)
            n += 2**k
    return res

def _trimStruct2D(s):
    """
    remove highest frequencies from structured 2D trasnform
    """
    res = {}; res.update(s)
    N = s['N']
    for x in ['D', 'H', 'V']:
        res.pop('f%i_%s'%(N-1, x))
        res.pop('i%i_%s'%(N-1, x))
    res['N']-=1
    return res

def _expandStruct2D(s):
    """
    remove highest frequencies from structured 2D trasnform
    """
    res = {}; res.update(s)
    N = s['N']
    res['f%i_D'%(N-1)] = np.zeros((2**N, 2**N))
    res['i%i_D'%(N-1)] = (2**(N-1), 2**N, 2**(N-1), 2**N)
    res['f%i_V'%(N-1)] = np.zeros((2**N, 2**N))
    res['i%i_V'%(N-1)] = (0, 2**(N-1), 2**(N-1), 2**N)
    res['f%i_H'%(N-1)] = np.zeros((2**N, 2**N))
    res['i%i_H'%(N-1)] = (2**(N-1), 2**N, 0, 2**(N-1))
    res['N']+=1
    return res

def structInv2D(x, order=4, retT=False):
    """
    structured inverse 2D transform.

    x is a dictionnary with separate frequencies and indexes for
    reconstruction (see 'struct2D'). """
    #N = max([int(k[1:].split('_')[0]) for k in x.keys()])+1
    N = x['N']
    res = np.zeros((2**N, 2**N))
    # -- reconstruct 2D transform from dict:
    for i in filter(lambda k: k.startswith('i'), x.keys()):
        #print(i, x[i])
        res[x[i][0]:x[i][1], x[i][2]:x[i][3]] = x['f'+i[1:]]
    if retT:
        return res
    else:
        #print('N=', N, 'o=', order)
        return twoD(res, -N, order=order)

def sparse2Dfrac(x, order=4, frac=10, retT=False):
    """
    x is a 2D image of shape (2**N, 2**N) shape.
    keep fraction "frac" in percent of each frequencies
    """
    S = struct2D(x, order=order)
    actual_frac = 0.
    for i in filter(lambda k: k.startswith('f'), S.keys()):
        o = float(i.split('_')[0][1:])
        if o>0:
            f = 1-(frac/100.)**(o*1./S['N'])
            #print(f)
            mask = np.abs(S[i])>=np.percentile(np.abs(S[i]), 100*f)
            S[i] *= mask
    return structInv2D(S, order=order, retT=retT)

def sparse2Dsigma(x, order=4, nsigma=10, retT=False):
    """
    x is a 2D image of shape (2**N, 2**N) shape.
    keep fraction "frac" in percent of each frequencies
    """
    S = struct2D(x, order=order)
    for i in filter(lambda k: k.startswith('f'), S.keys()):
        if float(i.split('_')[0][1:])>0:
            sigma = 0.5*(np.percentile(S[i], 84)-
                         np.percentile(S[i], 16))
            mask = np.abs(S[i])>=nsigma*sigma
            S[i] *= mask
    return structInv2D(S, order=order, retT=retT)

def psd2D(x):
    """
    x = 2D (image)
    returns: ?
    """
    N = int(np.log(max(x.shape))/np.log(2))+1
    tmp = np.zeros((2**N,2**N))
    # -- zero padding:
    tmp[:x.shape[0],:x.shape[1]]=x

    dwt = twoD(tmp,N)
    res = []
    for k in range(N+1):
        filt = np.zeros(N+1)
        filt[k]=1.0
        filt2D = build2DFilt(2**N,filt)
        res.append((dwt**2*filt2D).sum()/filt2D.sum())
    return res



import numpy as np

def _testT3formula(D, tri, key='OI_VIS'):
    tmp = {}
    s, t, w0, w1, w2 = D['OI_T3'][tri]['formula']
    keys = [k for k in ['OI_VIS', 'OI_VIS2'] if k in D]
    for key in keys:
        tmp[key+' mjd0'] = D['OI_T3'][tri]['MJD']-D[key][t[0]]['MJD'][w0]
        tmp[key+' u0'] = D['OI_T3'][tri]['u1']-s[0]*D[key][t[0]]['u'][w0]
        tmp[key+' v0'] = D['OI_T3'][tri]['v1']-s[0]*D[key][t[0]]['v'][w0]

        tmp[key+' mjd1'] = D['OI_T3'][tri]['MJD']-D[key][t[1]]['MJD'][w1]
        tmp[key+' u1'] = D['OI_T3'][tri]['u2']-s[1]*D[key][t[1]]['u'][w1]
        tmp[key+' v1'] = D['OI_T3'][tri]['v2']-s[1]*D[key][t[1]]['v'][w1]

        tmp[key+' mjd2'] = D['OI_T3'][tri]['MJD']-D[key][t[2]]['MJD'][w2]
        tmp[key+' u2'] = -D['OI_T3'][tri]['u1']-D['OI_T3'][tri]['u2']-s[2]*D[key][t[2]]['u'][w2]
        tmp[key+' v2'] = -D['OI_T3'][tri]['v1']-D['OI_T3'][tri]['v2']-s[2]*D[key][t[2]]['v'][w2]
    return tmp

def checkT3formula(oi, tol=1e-9):
    """
    oi is a PMOIRED object. Test if the OI_T3 formula are correct. That is, it
    checks if each OI_T3 is proprly associated

    """
    test = {} # keyed by filename
    for D in oi.data:
        if 'OI_T3' in D:
            test[D['filename']] = {}
            for tri in D['OI_T3']:
                test[D['filename']][tri] = _testT3formula(D, tri)
    try:
        for i,D in enumerate(oi._merged):
            if 'OI_T3' in D:
                test['MERGED'+str(i)] = {}
                for tri in D['OI_T3']:
                    test['MERGED'+str(i)][tri] = _testT3formula(D, tri)
    except:
        print('no merged data')
        pass

    err, msg = False, []
    for F in test: # for each filename
        for tri in test[F]: # for each triangle
            for k in test[F][tri]:
                if any(np.abs(test[F][tri][k])>tol):
                    msg += [(F, tri, k, np.abs(test[F][tri][k]))]
                    err = True
    return not err, msg

def checkShapes(oi):
    test = {}
    for i,D in enumerate(oi.data):
        test[D['filename']] = {}
        nWL = len(D['WL'])
        for e in ['OI_VIS', 'OI_VIS2', 'OI_T3']:
            for k in D[e]:
                nMJD = len(D[e][k]['MJD'])
                for x in D[e][k]:
                    if type(D[e][k][x]) == np.ndarray:
                        s = D[e][k][x].shape
                        #print(s, nWL, nMJD)
                        if len(s)==1 and s[0]==nMJD:
                            pass
                        elif len(s)==2 and s[0]==nMJD and s[1]==nWL:
                            pass
                        else:
                            test[D['filename']][e+k+x] = (s, nWL, nMJD)
    return test
